#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from collections.abc import Callable
import functools
import json
from typing import Any, cast

import jsonschema
import yaml

from ironicclient import exc
from ironicclient.v1 import client as v1_client

_CREATE_SCHEMA: dict[str, object] = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Schema for ironic resources file",
    "type": "object",
    "properties": {
        "chassis": {
            "type": "array",
            "items": {
                "type": "object"
            }
        },
        "nodes": {
            "type": "array",
            "items": {
                "type": "object"
            }
        }
    },
    "additionalProperties": False
}


def create_resources(
    client: v1_client.Client,
    filenames: list[str],
) -> None:
    """Create resources using their JSON or YAML descriptions.

    :param client: an instance of ironic client;
    :param filenames: a list of filenames containing JSON or YAML resources
        definitions.
    :raises: ClientException if any operation during files processing/resource
        creation fails.
    """
    errors: list[Exception] = []
    resources: list[dict[str, object]] = []
    for resource_file in filenames:
        try:
            resource = load_from_file(resource_file)
            jsonschema.validate(resource, _CREATE_SCHEMA)
            resources.append(resource)
        except (exc.ClientException, jsonschema.ValidationError) as e:
            errors.append(e)
    if errors:
        raise exc.ClientException('While validating the resources file(s), the'
                                  ' following error(s) were encountered:\n%s' %
                                  '\n'.join(str(e) for e in errors))
    for r in resources:
        errors.extend(create_chassis(
            client,
            cast(list[dict[str, object]], r.get('chassis', []))))
        errors.extend(create_nodes(
            client,
            cast(list[dict[str, object]], r.get('nodes', []))))
    if errors:
        raise exc.ClientException('During resources creation, the following '
                                  'error(s) were encountered:\n%s' %
                                  '\n'.join(str(e) for e in errors))


def load_from_file(filename: str) -> dict[str, object]:
    """Deserialize JSON or YAML from file.

    :param filename: name of the file containing JSON or YAML.
    :returns: a dictionary deserialized from JSON or YAML.
    :raises: ClientException if the file can not be loaded or if its contents
        is not a valid JSON or YAML, or if the file extension is not supported.
    """
    try:
        with open(filename) as f:
            if filename.endswith('.yaml'):
                return cast(dict[str, object], yaml.safe_load(f))
            elif filename.endswith('.json'):
                return cast(dict[str, object], json.load(f))
            else:
                # The file is neither .json, nor .yaml, raise an exception
                raise exc.ClientException(
                    'Cannot process file "%(file)s" - it must have .json or '
                    '.yaml extension.' % {'file': filename})
    except IOError as e:
        raise exc.ClientException('Cannot read file "%(file)s" due to '
                                  'error: %(err)s' %
                                  {'err': e, 'file': filename})
    except (ValueError, yaml.YAMLError) as e:
        # json.load raises only ValueError
        raise exc.ClientException('File "%(file)s" is invalid due to error: '
                                  '%(err)s' % {'err': e, 'file': filename})


def create_single_handler(
    resource_type: str,
) -> Callable[
    [Callable[..., str]],
    Callable[..., tuple[str | None, Exception | None]],
]:
    """Catch errors of the creation of a single resource.

    This decorator appends an error (which is an instance of some client
    exception class) to the return value of the create_method, changing the
    return value from just UUID to (UUID, error), and does some exception
    handling.

    :param resource_type: string value, the type of the resource being created,
        e.g. 'node', used purely for exception messages.
    """

    def outer_wrapper(
        create_method: Callable[..., str],
    ) -> Callable[..., tuple[str | None, Exception | None]]:
        @functools.wraps(create_method)
        def wrapper(
            client: v1_client.Client,
            **params: object,
        ) -> tuple[str | None, Exception | None]:
            uuid: str | None = None
            error: Exception | None = None
            try:
                uuid = create_method(client, **params)
            except exc.InvalidAttribute as e:
                error = exc.InvalidAttribute(
                    'Cannot create the %(resource)s with attributes '
                    '%(params)s. One or more attributes are invalid: %(err)s' %
                    {'params': params, 'resource': resource_type, 'err': e}
                )
            except Exception as e:
                error = exc.ClientException(
                    'Unable to create the %(resource)s with the specified '
                    'attributes: %(params)s. The error is: %(error)s' %
                    {'error': e, 'resource': resource_type, 'params': params})
            return uuid, error
        return wrapper
    return outer_wrapper


@create_single_handler('node')
def create_single_node(
    client: v1_client.Client, **params: Any,
) -> str:
    """Call the client to create a node.

    :param client: ironic client instance.
    :param params: dictionary to be POSTed to /nodes endpoint, excluding
        "ports" and "portgroups" keys.
    :returns: UUID of the created node or None in case of exception, and an
        exception, if it appears.
    :raises: InvalidAttribute, if some parameters passed to client's
        create_method are invalid.
    :raises: ClientException, if the creation of the node fails.
    """
    params.pop('ports', None)
    params.pop('portgroups', None)
    traits = params.pop('traits', None)
    ret = client.node.create(**params)
    if ret is None:
        raise exc.ClientException(
            'Unable to create the node: '
            'the API returned an empty response')
    if traits:
        client.node.set_traits(ret.uuid, cast(list[str], traits))
    return cast(str, ret.uuid)


@create_single_handler('port')
def create_single_port(
    client: v1_client.Client, **params: Any,
) -> str:
    """Call the client to create a port.

    :param client: ironic client instance.
    :param params: dictionary to be POSTed to /ports endpoint.
    :returns: UUID of the created port or None in case of exception, and an
        exception, if it appears.
    :raises: InvalidAttribute, if some parameters passed to client's
        create_method are invalid.
    :raises: ClientException, if the creation of the port fails.
    """
    ret = client.port.create(**params)
    if ret is None:
        raise exc.ClientException(
            'Unable to create the port: '
            'the API returned an empty response')
    return cast(str, ret.uuid)


@create_single_handler('port group')
def create_single_portgroup(
    client: v1_client.Client, **params: Any,
) -> str:
    """Call the client to create a port group.

    :param client: ironic client instance.
    :param params: dictionary to be POSTed to /portgroups endpoint, excluding
        "ports" key.
    :returns: UUID of the created port group or None in case of exception, and
        an exception, if it appears.
    :raises: InvalidAttribute, if some parameters passed to client's
        create_method are invalid.
    :raises: ClientException, if the creation of the portgroup fails.
    """
    params.pop('ports', None)
    ret = client.portgroup.create(**params)
    if ret is None:
        raise exc.ClientException(
            'Unable to create the port group: '
            'the API returned an empty response')
    return cast(str, ret.uuid)


@create_single_handler('chassis')
def create_single_chassis(
    client: v1_client.Client, **params: Any,
) -> str:
    """Call the client to create a chassis.

    :param client: ironic client instance.
    :param params: dictionary to be POSTed to /chassis endpoint, excluding
        "nodes" key.
    :returns: UUID of the created chassis or None in case of exception, and an
        exception, if it appears.
    :raises: InvalidAttribute, if some parameters passed to client's
        create_method are invalid.
    :raises: ClientException, if the creation of the chassis fails.
    """
    params.pop('nodes', None)
    ret = client.chassis.create(**params)
    if ret is None:
        raise exc.ClientException(
            'Unable to create the chassis: '
            'the API returned an empty response')
    return cast(str, ret.uuid)


def create_ports(
    client: v1_client.Client,
    port_list: list[dict[str, object]],
    node_uuid: str,
    portgroup_uuid: str | None = None,
) -> list[Exception]:
    """Create ports from dictionaries.

    :param client: ironic client instance.
    :param port_list: list of dictionaries to be POSTed to /ports
        endpoint.
    :param node_uuid: UUID of a node the ports should be associated with.
    :param portgroup_uuid: UUID of a port group the ports should be associated
        with, if they are its members.
    :returns: array of exceptions encountered during creation.
    """
    errors: list[Exception] = []
    for port in port_list:
        port_node_uuid = port.get('node_uuid')
        if port_node_uuid and port_node_uuid != node_uuid:
            errors.append(exc.ClientException(
                'Cannot create a port as part of node %(node_uuid)s '
                'because the port %(port)s has a different node UUID '
                'specified.',
                {'node_uuid': node_uuid,
                 'port': port}))
            continue
        port['node_uuid'] = node_uuid
        if portgroup_uuid:
            port_portgroup_uuid = port.get('portgroup_uuid')
            if port_portgroup_uuid and port_portgroup_uuid != portgroup_uuid:
                errors.append(exc.ClientException(
                    'Cannot create a port as part of port group '
                    '%(portgroup_uuid)s because the port %(port)s has a '
                    'different port group UUID specified.',
                    {'portgroup_uuid': portgroup_uuid,
                     'port': port}))
                continue
            port['portgroup_uuid'] = portgroup_uuid
        port_uuid, error = create_single_port(client, **port)
        if error:
            errors.append(error)
    return errors


def create_portgroups(
    client: v1_client.Client,
    portgroup_list: list[dict[str, object]],
    node_uuid: str,
) -> list[Exception]:
    """Create port groups from dictionaries.

    :param client: ironic client instance.
    :param portgroup_list: list of dictionaries to be POSTed to /portgroups
        endpoint, if some of them contain "ports" key, its content is POSTed
        separately to /ports endpoint.
    :param node_uuid: UUID of a node the port groups should be associated with.
    :returns: array of exceptions encountered during creation.
    """
    errors: list[Exception] = []
    for portgroup in portgroup_list:
        portgroup_node_uuid = portgroup.get('node_uuid')
        if portgroup_node_uuid and portgroup_node_uuid != node_uuid:
            errors.append(exc.ClientException(
                'Cannot create a port group as part of node %(node_uuid)s '
                'because the port group %(portgroup)s has a different node '
                'UUID specified.',
                {'node_uuid': node_uuid,
                 'portgroup': portgroup}))
            continue
        portgroup['node_uuid'] = node_uuid
        portgroup_uuid, error = create_single_portgroup(client, **portgroup)
        if error:
            errors.append(error)
        ports = portgroup.get('ports')
        # Port group UUID == None means that port group creation failed, don't
        # create the ports inside it
        if ports is not None and portgroup_uuid is not None:
            errors.extend(create_ports(
                client,
                cast(list[dict[str, object]], ports),
                node_uuid,
                portgroup_uuid=portgroup_uuid))
    return errors


def create_nodes(
    client: v1_client.Client,
    node_list: list[dict[str, object]],
    chassis_uuid: str | None = None,
) -> list[Exception]:
    """Create nodes from dictionaries.

    :param client: ironic client instance.
    :param node_list: list of dictionaries to be POSTed to /nodes
        endpoint, if some of them contain "ports" key, its content is POSTed
        separately to /ports endpoint.
    :param chassis_uuid: UUID of a chassis the nodes should be associated with.
    :returns: array of exceptions encountered during creation.
    """
    errors: list[Exception] = []
    for node in node_list:
        if chassis_uuid is not None:
            node_chassis_uuid = node.get('chassis_uuid')
            if node_chassis_uuid and node_chassis_uuid != chassis_uuid:
                errors.append(exc.ClientException(
                    'Cannot create a node as part of chassis %(chassis_uuid)s '
                    'because the node %(node)s has a different chassis UUID '
                    'specified.' %
                    {'chassis_uuid': chassis_uuid,
                     'node': node}))
                continue
            node['chassis_uuid'] = chassis_uuid
        node_uuid, error = create_single_node(client, **node)
        if error:
            errors.append(error)
        ports = node.get('ports')
        portgroups = node.get('portgroups')
        # Node UUID == None means that node creation failed, don't
        # create the port(group)s inside it
        if node_uuid is not None:
            if portgroups is not None:
                errors.extend(create_portgroups(
                    client,
                    cast(list[dict[str, object]], portgroups),
                    node_uuid))
            if ports is not None:
                errors.extend(create_ports(
                    client,
                    cast(list[dict[str, object]], ports),
                    node_uuid))
    return errors


def create_chassis(
    client: v1_client.Client,
    chassis_list: list[dict[str, object]],
) -> list[Exception]:
    """Create chassis from dictionaries.

    :param client: ironic client instance.
    :param chassis_list: list of dictionaries to be POSTed to /chassis
        endpoint, if some of them contain "nodes" key, its content is POSTed
        separately to /nodes endpoint.
    :returns: array of exceptions encountered during creation.
    """
    errors: list[Exception] = []
    for chassis in chassis_list:
        chassis_uuid, error = create_single_chassis(client, **chassis)
        if error:
            errors.append(error)
        nodes = chassis.get('nodes')
        # Chassis UUID == None means that chassis creation failed, don't
        # create the nodes inside it
        if nodes is not None and chassis_uuid is not None:
            errors.extend(create_nodes(
                client,
                cast(list[dict[str, object]], nodes),
                chassis_uuid=chassis_uuid))
    return errors



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import copy

from ironicclient.osc.v1 import baremetal_shard
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestBaremetalShard(baremetal_fakes.TestBaremetal):
    def setUp(self) -> None:
        super(TestBaremetalShard, self).setUp()
        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()


class TestBaremetalShardList(TestBaremetalShard):
    def setUp(self) -> None:
        super(TestBaremetalShardList, self).setUp()

        # Return a list containing mocked shard data to
        # simulate the expected output
        self.baremetal_mock.shard.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.SHARD),
                loaded=True,
            ),
        ]

        # Get the command object to test
        self.cmd = baremetal_shard.ListBaremetalShard(self.app, None)

    def test_shard_list(self) -> None:
        arglist = []
        verifylist = []

        # Parse arguments and invoke command
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Define expected columns and data output
        collist = ("name", "count")
        datalist = ((baremetal_fakes.baremetal_shard_name,
                     baremetal_fakes.baremetal_shard_count), )

        self.assertEqual(collist, columns)
        self.assertEqual(datalist, tuple(data))



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from collections.abc import Callable
import logging
import os
from typing import Any, cast

from oslo_utils import strutils

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.v1 import port
from ironicclient.v1 import volume_connector
from ironicclient.v1 import volume_target

_power_states: dict[str, str] = {
    'on': 'power on',
    'off': 'power off',
    'reboot': 'rebooting',
    'soft off': 'soft power off',
    'soft reboot': 'soft rebooting',
}


LOG: logging.Logger = logging.getLogger(__name__)
_DEFAULT_POLL_INTERVAL: int = 2


class Node(base.Resource):
    def __repr__(self) -> str:
        return "<Node %s>" % self._info


class NodeManager(base.CreateManager[Node]):
    resource_class: type[Node] = Node
    _creation_attributes: list[str] = [
        'chassis_uuid', 'driver', 'driver_info',
        'extra', 'uuid', 'properties', 'name',
        'bios_interface', 'boot_interface',
        'console_interface', 'deploy_interface',
        'disable_power_off', 'inspect_interface',
        'management_interface', 'network_interface',
        'power_interface', 'raid_interface',
        'rescue_interface', 'storage_interface',
        'vendor_interface', 'firmware_interface',
        'resource_class', 'conductor_group',
        'automated_clean', 'network_data', 'parent_node',
        'owner', 'lessee', 'shard', 'description',
        'instance_name',
    ]
    _resource_name: str = 'nodes'

    def list_ports(
        self,
        node_id: str,
        marker: str | None = None,
        limit: int | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[port.Port]:
        """List all the ports for a given node.

        :param node_id: Name or UUID of the node.
        :param marker: Optional, the UUID of a port, eg the last
                       port from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of ports to return.
            2) limit == 0, return the entire list of ports.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about ports.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of ports.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(_("Can't fetch a subset of fields "
                                         "with 'detail' set"))

        filters = utils.common_filters(marker, limit, sort_key, sort_dir,
                                       fields)

        path = "%s/ports" % node_id
        if detail:
            path += '/detail'

        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path), "ports",
                obj_class=port.Port,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path), "ports",
                obj_class=port.Port, limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )

    def list_volume_connectors(
        self,
        node_id: str,
        marker: str | None = None,
        limit: int | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[volume_connector.VolumeConnector]:
        """List all the volume connectors for a given node.

        :param node_id: Name or UUID of the node.
        :param marker: Optional, the UUID of a volume connector, eg the last
                       volume connector from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of volume connectors to return.
            2) limit == 0, return the entire list of volume connectors.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about volume connectors.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of volume connectors.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(_("Can't fetch a subset of fields "
                                         "with 'detail' set"))

        filters = utils.common_filters(marker=marker, limit=limit,
                                       sort_key=sort_key, sort_dir=sort_dir,
                                       fields=fields, detail=detail)

        path = "%s/volume/connectors" % node_id
        if filters:
            path += '?' + '&'.join(filters)
        vc = volume_connector.VolumeConnector
        if limit is None:
            return self._list(
                self._path(path), response_key="connectors",
                obj_class=vc,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path), response_key="connectors",
                limit=limit,
                obj_class=vc,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )

    def list_volume_targets(
        self,
        node_id: str,
        marker: str | None = None,
        limit: int | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[volume_target.VolumeTarget]:
        """List all the volume targets for a given node.

        :param node_id: Name or UUID of the node.
        :param marker: Optional, the UUID of a volume target, eg the last
                       volume target from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of volume targets to return.
            2) limit == 0, return the entire list of volume targets.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about volume targets.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of volume targets.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(_("Can't fetch a subset of fields "
                                         "with 'detail' set"))

        filters = utils.common_filters(marker=marker, limit=limit,
                                       sort_key=sort_key, sort_dir=sort_dir,
                                       fields=fields, detail=detail)

        path = "%s/volume/targets" % node_id
        if filters:
            path += '?' + '&'.join(filters)
        vt = volume_target.VolumeTarget
        if limit is None:
            return self._list(
                self._path(path), response_key="targets",
                obj_class=vt,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path), response_key="targets",
                limit=limit,
                obj_class=vt,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )

    def list_children_of_node(
        self,
        node_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[str]:
        """Get a list of child nodes for the supplied node_id.

        :param node_id: The name or UUID of a node.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of UUIDs representing child nodes for the supplied
                  node_id..
        """
        path = "%s/children" % node_id
        return cast(list[str], self._list_primitives(
            self._path(path), "children",
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        ))

    def get(
        self,
        node_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        return self._get(resource_id=node_id, fields=fields,
                         os_ironic_api_version=os_ironic_api_version,
                         global_request_id=global_request_id)

    def get_by_instance_uuid(
        self,
        instance_uuid: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node:
        path = '?instance_uuid=%s' % instance_uuid
        if fields is not None:
            path += '&fields=' + ','.join(fields)
        else:
            path = 'detail' + path

        nodes = self._list(
            self._path(path), 'nodes',
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)
        # get all the details of the node assuming that
        # filtering by instance_uuid returns a collection
        # of one node if successful.
        if len(nodes) == 1:
            return nodes[0]
        else:
            raise exc.NotFound()

    def delete(
        self,
        node_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        self._delete(
            resource_id=node_id,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def update(
        self,
        node_id: str,
        patch: list[dict[str, Any]] | dict[str, Any] | None,
        http_method: str = 'PATCH',
        os_ironic_api_version: str | None = None,
        reset_interfaces: bool | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        params: dict[str, bool] = {}
        if reset_interfaces is not None:
            params['reset_interfaces'] = reset_interfaces
        return self._update(
            resource_id=node_id,
            patch=patch,  # type: ignore[arg-type]
            method=http_method,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
            params=params,
        )

    def vendor_passthru(
        self,
        node_id: str,
        method: str,
        args: dict[str, Any] | None = None,
        http_method: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Issue requests for vendor-specific actions on a given node.

        :param node_id: The UUID of the node.
        :param method: Name of the vendor method.
        :param args: Optional. The arguments to be passed to the method.
        :param http_method: The HTTP method to use on the request.
                            Defaults to POST.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        if args is None:
            args = {}

        if http_method is None:
            http_method = 'POST'

        http_method = http_method.upper()

        path = "%s/vendor_passthru/%s" % (node_id, method)
        if http_method in ('POST', 'PUT', 'PATCH'):
            return self.update(
                path, args, http_method=http_method,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        elif http_method == 'DELETE':
            self.delete(
                path,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
            return None
        elif http_method == 'GET':
            return self.get(
                path,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            raise exc.InvalidAttribute(
                _('Unknown HTTP method: %s') % http_method)

    def vif_list(
        self,
        node_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Node]:
        """List VIFs attached to a given node.

        :param node_ident: The UUID or Name of the node.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/vifs" % node_ident

        return self._list(self._path(path), "vifs",
                          os_ironic_api_version=os_ironic_api_version,
                          global_request_id=global_request_id)

    def vif_attach(
        self,
        node_ident: str,
        vif_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        **kwargs: str,
    ) -> None:
        """Attach VIF to a given node.

        :param node_ident: The UUID or Name of the node.
        :param vif_id: The UUID or Name of the VIF to attach.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :param kwargs: A dictionary containing the attributes of the resource
                       that will be created.
        """
        path = "%s/vifs" % node_ident
        data: dict[str, str] = {"id": vif_id}
        if 'id' in kwargs:
            raise exc.InvalidAttribute("The attribute 'id' can't be "
                                       "specified in vif-info")
        data.update(kwargs)
        # TODO(vdrok): cleanup places doing custom path and http_method
        self.update(path, data, http_method="POST",
                    os_ironic_api_version=os_ironic_api_version,
                    global_request_id=global_request_id)

    def vif_detach(
        self,
        node_ident: str,
        vif_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Detach VIF from a given node.

        :param node_ident: The UUID or Name of the node.
        :param vif_id: The UUID or Name of the VIF to detach.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/vifs/%s" % (node_ident, vif_id)
        self.delete(path, os_ironic_api_version=os_ironic_api_version,
                    global_request_id=global_request_id)

    def set_maintenance(
        self,
        node_id: str,
        state: bool | str,
        maint_reason: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Set the maintenance mode for the node.

        :param node_id: The UUID of the node.
        :param state: the maintenance mode; either a Boolean or a string
                      representation of a Boolean (eg, 'true', 'on', 'false',
                      'off'). True to put the node in maintenance mode; False
                      to take the node out of maintenance mode.
        :param maint_reason: Optional string. Reason for putting node
                             into maintenance mode.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :raises: InvalidAttribute if state is an invalid string (that doesn't
                 represent a Boolean).
        """
        if isinstance(state, bool):
            maintenance_mode = state
        else:
            try:
                maintenance_mode = strutils.bool_from_string(state, True)
            except ValueError as e:
                raise exc.InvalidAttribute(_("Argument 'state': %(err)s") %
                                           {'err': e})
        path = "%s/maintenance" % node_id
        if maintenance_mode:
            reason = {'reason': maint_reason}
            return self.update(
                path, reason, http_method='PUT',
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            self.delete(
                path,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
            return None

    def set_power_state(
        self,
        node_id: str,
        state: str,
        soft: bool = False,
        timeout: int | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Sets power state for a node.

        :param node_id: Node identifier
        :param state: One of target power state, 'on', 'off', or 'reboot'
        :param soft: The flag for graceful power 'off' or 'reboot'
        :param timeout: The timeout (in seconds) positive integer value (> 0)
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :raises: ValueError if 'soft' or 'timeout' option is invalid
        :returns: The status of the request
        """
        if state == 'on' and soft:
            raise ValueError(
                _("'soft' option is invalid for the power-state 'on'"))

        path = "%s/states/power" % node_id

        requested_state = 'soft ' + state if soft else state
        target = _power_states.get(requested_state, state)

        body: dict[str, str | int] = {'target': target}
        if timeout is not None:
            msg = _("'timeout' option for setting power state must have "
                    "positive integer value (> 0)")
            try:
                timeout = int(timeout)
            except (ValueError, TypeError):
                raise ValueError(msg)

            if timeout <= 0:
                raise ValueError(msg)
            body = {'target': target, 'timeout': timeout}

        return self.update(path, body, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def set_boot_mode(
        self,
        node_id: str,
        state: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Sets boot mode for a node.

        :param node_id: Node identifier
        :param state: One of target boot modes, 'uefi' or 'bios'
        :param os_ironic_api_version: String version (e.g. "1.76") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :raises: ValueError if boot mode is not one of 'uefi' / 'bios'
        :returns: The status of the request
        """
        if state not in ('uefi', 'bios'):
            raise ValueError(
                _("Valid boot modes are 'uefi' or 'bios'"))

        path = "%s/states/boot_mode" % node_id
        target = state
        body = {'target': target}

        return self.update(path, body, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def set_secure_boot(
        self,
        node_id: str,
        state: bool | str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Set the secure boot state for the node.

        :param node_id: The UUID of the node.
        :param state: the secure boot state; either a Boolean or a string
                      representation of a Boolean (eg, 'true', 'on', 'false',
                      'off'). True to turn secure boot on; False
                      to turn secure boot off.
        :param os_ironic_api_version: String version (e.g. "1.76") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :raises: InvalidAttribute if state is an invalid string (that doesn't
                 represent a Boolean).
        """
        if isinstance(state, bool):
            target = state
        else:
            try:
                target = strutils.bool_from_string(state, strict=True)
            except ValueError as e:
                raise exc.InvalidAttribute(_("Argument 'state': %(err)s") %
                                           {'err': e})
        path = "%s/states/secure_boot" % node_id
        body = {'target': target}

        return self.update(path, body, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def set_target_raid_config(
        self,
        node_ident: str,
        target_raid_config: dict[str, Any],
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Sets target_raid_config for a node.

        :param node_ident: Node identifier
        :param target_raid_config: A dictionary with the target RAID
            configuration; may be empty.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: status of the request
        """
        path = "%s/states/raid" % node_ident
        return self.update(path, target_raid_config, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def validate(
        self,
        node_uuid: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        path = "%s/validate" % node_uuid
        return self.get(path, os_ironic_api_version=os_ironic_api_version,
                        global_request_id=global_request_id)

    def set_provision_state(
        self,
        node_uuid: str,
        state: str,
        configdrive: str | dict[str, Any] | list[Any] | bytes | None = None,
        cleansteps: list[dict[str, Any]] | None = None,
        rescue_password: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        deploysteps: list[dict[str, Any]] | None = None,
        servicesteps: list[dict[str, Any]] | None = None,
        runbook: str | None = None,
        disable_ramdisk: bool | str | None = None,
    ) -> Node | None:
        """Set the provision state for the node.

        :param node_uuid: The UUID or name of the node.
        :param state: The desired provision state. One of 'active', 'deleted',
             'rebuild', 'inspect', 'provide', 'manage', 'clean', 'abort',
             'rescue', 'unrescue'.
        :param configdrive: One of:

            * a gzipped, base64-encoded configuration drive string
            * a dictionary to build config drive from
            * a path to the configuration drive file (ISO 9660 or VFAT)
            * a path to a directory containing the config drive files
            * a path to a JSON file to build config from

            In case it's a directory, a config drive will be generated from
            it. In case it's a dictionary or a JSON file, a config drive will
            be generated on the server side (requires API version 1.56).
            This is only valid when setting state to 'active'.
        :param cleansteps: The clean steps as a list of clean-step
            dictionaries; each dictionary should have keys 'interface' and
            'step', and optional key 'args'. This must be specified (and is
            only valid) when setting provision-state to 'clean'.
        :param rescue_password: A string to be used as the login password
            inside the rescue ramdisk once a node is rescued. This must be
            specified (and is only valid) when setting 'state' to 'rescue'.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :param deploysteps: The deploy steps as a list of deploy-step
            dictionaries; each dictionary should have keys 'interface', 'step',
            'priority', and optional key 'args'. This is optional and is
            only valid when setting provision-state to 'active' or 'rebuild'.
        :param servicesteps: The service steps as list of service-step
            dictionaries; each dictionary should have keys 'interface', 'step',
            and optional key 'args' when setting an 'active' nodes to
            'service'.
        :param runbook: The identifier of a predefined runbook to use for
            provisioning.
        :param disable_ramdisk: Boolean if set to true will not boot the
            ironic-python-agent for cleaning. Only valid when setting 'state'
            to 'clean' or 'service' and only for steps explicitly marked as
            not requiring the ironic-python-agent can use this.
        :raises: InvalidAttribute if there was an error with the clean steps or
            deploy steps
        :returns: The status of the request
        """

        path = "%s/states/provision" % node_uuid
        body: dict[str, Any] = {'target': state}
        if configdrive:
            if isinstance(configdrive, str):
                if os.path.isfile(configdrive):
                    with open(configdrive, 'rb') as f:
                        configdrive = f.read()
                    json_data = utils.get_json_data(configdrive)
                    if json_data is not None:
                        configdrive = json_data
                elif os.path.isdir(configdrive):
                    configdrive = utils.make_configdrive(configdrive)
                else:
                    raise ValueError('Config drive seems to refer to a file '
                                     'or directory but this file/directory '
                                     'does not exist: %s.' % configdrive)

            if isinstance(configdrive, bytes):
                try:
                    configdrive = configdrive.decode('utf-8')
                except UnicodeError:
                    raise ValueError('Config drive must be a dictionary or '
                                     'a base64 encoded string')
            body['configdrive'] = configdrive
        elif cleansteps:
            body['clean_steps'] = cleansteps
        elif rescue_password:
            body['rescue_password'] = rescue_password

        if deploysteps:
            body['deploy_steps'] = deploysteps

        if servicesteps:
            body['service_steps'] = servicesteps

        if runbook:
            body['runbook'] = runbook

        if isinstance(disable_ramdisk, bool):
            body['disable_ramdisk'] = disable_ramdisk
        elif disable_ramdisk:
            try:
                body['disable_ramdisk'] = strutils.bool_from_string(
                    disable_ramdisk, True)
            except ValueError as e:
                raise exc.InvalidAttribute(_("Argument 'disable_ramdisk': "
                                             "%(err)s") % {'err': e})

        return self.update(path, body, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def states(
        self,
        node_uuid: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        path = "%s/states" % node_uuid
        return self.get(path, os_ironic_api_version=os_ironic_api_version,
                        global_request_id=global_request_id)

    def get_console(
        self,
        node_uuid: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, bool | dict[str, str] | None]:
        path = "%s/states/console" % node_uuid
        return cast(
            dict[str, bool | dict[str, str] | None],
            self._get_as_dict(
                path, os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def set_console_mode(
        self,
        node_uuid: str,
        enabled: bool | str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Set the console mode for the node.

        :param node_uuid: The UUID of the node.
        :param enabled: Either a Boolean or a string representation of a
                        Boolean. True to enable the console; False to disable.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/states/console" % node_uuid
        target = {'enabled': enabled}
        return self.update(path, target, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def set_boot_device(
        self,
        node_uuid: str,
        boot_device: str,
        persistent: bool = False,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        path = "%s/management/boot_device" % node_uuid
        target = {'boot_device': boot_device, 'persistent': persistent}
        return self.update(path, target, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def get_boot_device(
        self,
        node_uuid: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, str | bool]:
        path = "%s/management/boot_device" % node_uuid
        return cast(
            dict[str, str | bool],
            self._get_as_dict(
                path, os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def inject_nmi(
        self,
        node_uuid: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        path = "%s/management/inject_nmi" % node_uuid
        return self.update(path, {}, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def get_supported_boot_devices(
        self,
        node_uuid: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, list[str]]:
        path = "%s/management/boot_device/supported" % node_uuid
        return cast(
            dict[str, list[str]],
            self._get_as_dict(
                path, os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def get_vendor_passthru_methods(
        self,
        node_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, dict[str, bool | str | list[str]]]:
        path = "%s/vendor_passthru/methods" % node_ident
        return cast(
            dict[str, dict[str, bool | str | list[str]]],
            self._get_as_dict(
                path, os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def get_traits(
        self,
        node_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[str]:
        """Get traits for a node.

        :param node_ident: node UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/traits" % node_ident
        return cast(list[str], self._list_primitives(
            self._path(path), 'traits',
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id))

    def add_trait(
        self,
        node_ident: str,
        trait: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Add a trait to a node.

        :param node_ident: node UUID or name.
        :param trait: trait to add to the node.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/traits/%s" % (node_ident, trait)
        return self.update(path, None, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def set_traits(
        self,
        node_ident: str,
        traits: list[str],
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Node | None:
        """Set traits for a node.

        Removes any existing traits and adds the traits passed in to this
        method.

        :param node_ident: node UUID or name.
        :param traits: list of traits to add to the node.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/traits" % node_ident
        body = {'traits': traits}
        return self.update(path, body, http_method='PUT',
                           os_ironic_api_version=os_ironic_api_version,
                           global_request_id=global_request_id)

    def remove_trait(
        self,
        node_ident: str,
        trait: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Remove a trait from a node.

        :param node_ident: node UUID or name.
        :param trait: trait to remove from the node.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/traits/%s" % (node_ident, trait)
        self.delete(path, os_ironic_api_version=os_ironic_api_version,
                    global_request_id=global_request_id)

    def remove_all_traits(
        self,
        node_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Remove all traits from a node.

        :param node_ident: node UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/traits" % node_ident
        self.delete(path, os_ironic_api_version=os_ironic_api_version,
                    global_request_id=global_request_id)

    def get_bios_setting(
        self,
        node_ident: str,
        name: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, str] | None:
        """Get a BIOS setting from a node.

        :param node_ident: node UUID or name.
        :param name: BIOS setting name to get from the node.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/bios/%s" % (node_ident, name)
        return cast(
            dict[str, str] | None,
            self._get_as_dict(
                path, os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id).get(name),
        )

    def list_bios_settings(
        self,
        node_ident: str,
        detail: bool = False,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[dict[str, str | int | bool | list[str] | None]]:
        """List all BIOS settings from a node.

        :param node_ident: node UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :param detail: Optional, boolean whether to return detailed information
                       about bios settings.
        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        """
        if detail and fields:
            raise exc.InvalidAttribute(_("Can't fetch a subset of fields "
                                         "with 'detail' set"))

        filters = utils.common_filters(detail=detail, fields=fields)
        path = "%s/bios" % node_ident

        if filters:
            path += '?' + '&'.join(filters)

        return cast(
            list[dict[str, str | int | bool | list[str] | None]],
            self._list_primitives(
                self._path(path), 'bios',
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def _check_one_provision_state(
        self,
        node_ident: str,
        expected_state: str,
        fail_on_unexpected_state: bool = True,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> bool:
        # TODO(dtantsur): use version negotiation to request API 1.8 and use
        # the "fields" argument to reduce amount of data sent.
        node = self.get(
            node_ident, os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)
        if node is None:
            return False
        if node.provision_state == expected_state:
            LOG.debug('Node %(node)s reached provision state %(state)s',
                      {'node': node_ident, 'state': expected_state})
            return True

        # Note that if expected_state == 'error' we still succeed
        if (node.provision_state == 'error'
                or node.provision_state.endswith(' failed')):
            raise exc.StateTransitionFailed(
                _('Node %(node)s failed to reach state %(state)s. '
                    'It\'s in state %(actual)s, and has error: %(error)s') %
                {'node': node_ident, 'state': expected_state,
                    'actual': node.provision_state, 'error': node.last_error})

        if fail_on_unexpected_state and not node.target_provision_state:
            raise exc.StateTransitionFailed(
                _('Node %(node)s failed to reach state %(state)s. '
                    'It\'s in unexpected stable state %(actual)s') %
                {'node': node_ident, 'state': expected_state,
                 'actual': node.provision_state})

        LOG.debug('Still waiting for node %(node)s to reach state '
                  '%(state)s, the current state is %(actual)s',
                  {'node': node_ident, 'state': expected_state,
                   'actual': node.provision_state})

        return False

    def wait_for_provision_state(
        self,
        node_ident: str | list[str],
        expected_state: str,
        timeout: int | float = 0,
        poll_interval: int | float = _DEFAULT_POLL_INTERVAL,
        poll_delay_function: Callable[[int | float], object] | None = None,
        fail_on_unexpected_state: bool = True,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Helper function to wait for nodes to reach a given state.

        Polls Ironic API in a loop until node gets to a requested state.

        Fails in the following cases:
        * Timeout (if provided) is reached
        * Node's last_error gets set to a non-empty value
        * Unexpected stable state is reached and fail_on_unexpected_state is on
        * Error state is reached (if it's not equal to expected_state)

        :param node_ident: node UUID or name (one or a list)
        :param expected_state: expected final provision state
        :param timeout: timeout in seconds, no timeout if 0
        :param poll_interval: interval in seconds between 2 poll
        :param poll_delay_function: function to use to wait between polls
            (defaults to time.sleep). Should take one argument - delay time
            in seconds. Any exceptions raised inside it will abort the wait.
        :param fail_on_unexpected_state: whether to fail if the nodes
            reaches a different stable state.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :raises: StateTransitionFailed if node reached an error state
        :raises: StateTransitionTimeout on timeout
        """
        expected_state = expected_state.lower()
        if not isinstance(node_ident, list):
            node_ident = [node_ident]
        unfinished = node_ident

        def _timeout() -> str:
            return (
                _('Node(s) %(node)s failed to reach state %(state)s in '
                  '%(timeout)s seconds')
                % {'node': ', '.join(unfinished),
                   'state': expected_state,
                   'timeout': timeout}
            )

        for _count in utils.poll(timeout, poll_interval, poll_delay_function,
                                 _timeout):
            current, unfinished = unfinished, []
            for node in current:
                if not self._check_one_provision_state(
                        node,
                        expected_state,
                        fail_on_unexpected_state=fail_on_unexpected_state,
                        os_ironic_api_version=os_ironic_api_version,
                        global_request_id=global_request_id):
                    unfinished.append(node)
            if not unfinished:
                break

    def get_history_list(
        self,
        node_ident: str,
        detail: bool = False,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[dict[str, str]]:
        """Get node history event list.

        Provides the ability to query a node event history list from
        the API and return the API response to the caller.

        Requires API version 1.78.

        :param node_ident: The name or UUID of the node.
        :param detail: If detailed data should be returned in the
                       event list entry. Default False.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/history" % node_ident

        if detail:
            path = path + '?detail=%s' % detail

        return cast(
            list[dict[str, str]],
            self._list_primitives(
                self._path(path), 'history',
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def get_history_event(
        self,
        node_ident: str,
        event: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, str]:
        """Get a single event record for a node.

        Provides the ability to request, and return
        a node's single event history entry.

        :param node_ident: The name or UUID of the node.
        :param event: The UUID of the event entry as listed
                      in the node event history list.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/history/%s" % (node_ident, event)
        return cast(
            dict[str, str],
            self._get_as_dict(
                path, os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def get_inventory(
        self,
        node_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, Any]:
        """Get the hardware inventory of the node.

        Requires API version 1.81.

        :param node_ident: The name or UUID of the node.
        :param os_ironic_api_version: String version (e.g. "1.81") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/inventory" % node_ident
        return self._get_as_dict(
            path, os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def list_firmware_components(
        self,
        node_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[dict[str, str]]:
        """List all firmware components from a node.

        :param node_ident: node UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        path = "%s/firmware" % node_ident

        return cast(
            list[dict[str, str]],
            self._list_primitives(
                self._path(path), 'firmware',
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id),
        )

    def list(
        self,
        associated: bool | str | None = None,
        maintenance: bool | str | None = None,
        marker: str | None = None,
        limit: int | None = None,
        detail: bool = False,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        fields: list[str] | None = None,
        provision_state: str | None = None,
        driver: str | None = None,
        resource_class: str | None = None,
        chassis: str | None = None,
        fault: str | None = None,
        os_ironic_api_version: str | None = None,
        conductor_group: str | None = None,
        conductor: str | None = None,
        owner: str | None = None,
        retired: bool | str | None = None,
        lessee: str | None = None,
        shards: list[str] | None = None,
        sharded: bool | str | None = None,
        parent_node: str | None = None,
        include_children: bool | None = None,
        description_contains: str | None = None,
        global_request_id: str | None = None,
        instance_name: str | None = None,
    ) -> list[Node]:
        """Retrieve a list of nodes.

        :param associated: Optional. Either a Boolean or a string
                           representation of a Boolean that indicates whether
                           to return a list of associated (True or "True") or
                           unassociated (False or "False") nodes.
        :param maintenance: Optional. Either a Boolean or a string
                            representation of a Boolean that indicates whether
                            to return nodes in maintenance mode (True or
                            "True"), or not in maintenance mode (False or
                            "False").
        :param retired: Optional. Either a Boolean or a string representation
                        of a Boolean that indicates whether to return retired
                        nodes (True or "True").
        :param provision_state: Optional. String value to get only nodes in
                                that provision state.
        :param marker: Optional, the UUID of a node, eg the last
                       node from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of nodes to return.
            2) limit == 0, return the entire list of nodes.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param detail: Optional, boolean whether to return detailed information
                       about nodes.

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param driver: Optional. String value to get only nodes using that
                       driver.

        :param resource_class: Optional. String value to get only nodes
                               with the given resource class set.

        :param chassis: Optional, the UUID of a chassis. Used to get only
                        nodes of this chassis.

        :param fault: Optional. String value to get only nodes with
                      specified fault.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :param conductor_group: Optional. String value to get only nodes
                                with the given conductor group set.
        :param conductor: Optional. String value to get only nodes
                          mapped to the given conductor.
        :param owner: Optional. String value to get only nodes
                      mapped to a specific owner.
        :param lessee: Optional. String value to get only nodes
                       mapped to a specific lessee.
        :param shards: Optional. A list with a specified set of shards
                      to limit node returns to.
        :param sharded: Optional. Boolean value, when true get only nodes
                        with a non-null node.shard value, when false get only
                        nodes with a null node.shard value. None is a noop.
                        with a non-null node.shard value.
        :param parent_node: Optional. String value used to retrieve child
                            nodes with the supplied parent node.
        :param include_children: Optional. Boolean Value, only True is valid.
                                 Tells the ironic API to enumerate all child
                                 nodes which are normally hidden from the
                                 node list.
        :param description_contains: Optional. String value to get nodes
                                 with description contains specified value.
        :param instance_name: Optional. String value to get only nodes with
                               the given instance_name set.
        :returns: A list of nodes.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(_("Can't fetch a subset of fields "
                                         "with 'detail' set"))

        filters = utils.common_filters(marker, limit, sort_key, sort_dir,
                                       fields)
        if associated is not None:
            filters.append('associated=%s' % associated)
        if maintenance is not None:
            filters.append('maintenance=%s' % maintenance)
        if retired is not None:
            filters.append('retired=%s' % retired)
        if fault is not None:
            filters.append('fault=%s' % fault)
        if provision_state is not None:
            filters.append('provision_state=%s' % provision_state)
        if driver is not None:
            filters.append('driver=%s' % driver)
        if resource_class is not None:
            filters.append('resource_class=%s' % resource_class)
        if chassis is not None:
            filters.append('chassis_uuid=%s' % chassis)
        if conductor_group is not None:
            filters.append('conductor_group=%s' % conductor_group)
        if conductor is not None:
            filters.append('conductor=%s' % conductor)
        if owner is not None:
            filters.append('owner=%s' % owner)
        if lessee is not None:
            filters.append('lessee=%s' % lessee)
        if sharded is not None:
            filters.append('sharded=%s' % sharded)
        if shards is not None:
            filters.append('shard=%s' % ','.join(shards))
        if parent_node is not None:
            filters.append('parent_node=%s' % parent_node)
        if include_children:
            # NOTE(TheJulia): Only valid if True.
            filters.append('include_children=True')
        if description_contains is not None:
            filters.append('description_contains=%s' % description_contains)
        if instance_name is not None:
            filters.append('instance_name=%s' % instance_name)

        path = ''
        if detail:
            path += 'detail'
        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path), "nodes",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path), "nodes", limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import copy

import testtools
from testtools.matchers import HasLength

from ironicclient import exc
from ironicclient.tests.unit import utils
import ironicclient.v1.port

PORT = {'uuid': '11111111-2222-3333-4444-555555555555',
        'node_uuid': '55555555-4444-3333-2222-111111111111',
        'address': 'AA:BB:CC:DD:EE:FF',
        'pxe_enabled': True,
        'local_link_connection': {},
        'portgroup_uuid': '55555555-4444-3333-2222-111111111111',
        'physical_network': 'physnet1',
        'is_smartnic': False,
        'extra': {},
        'name': 'port_name'}

PORT2 = {'uuid': '55555555-4444-3333-2222-111111111111',
         'node_uuid': '55555555-4444-3333-2222-111111111111',
         'address': 'AA:AA:AA:BB:BB:BB',
         'pxe_enabled': True,
         'local_link_connection': {},
         'portgroup_uuid': '55555555-4444-3333-2222-111111111111',
         'physical_network': 'physnet2',
         'is_smartnic': True,
         'extra': {},
         'name': 'port2_name'}

CREATE_PORT = copy.deepcopy(PORT)
del CREATE_PORT['uuid']

CREATE_PORT_WITH_UUID = copy.deepcopy(PORT)

UPDATED_PORT = copy.deepcopy(PORT)
NEW_ADDR = 'AA:AA:AA:AA:AA:AA'
UPDATED_PORT['address'] = NEW_ADDR

fake_responses = {
    '/v1/ports':
    {
        'GET': (
            {},
            {"ports": [PORT]},
        ),
        'POST': (
            {},
            CREATE_PORT,
        ),
    },
    '/v1/ports/detail':
    {
        'GET': (
            {},
            {"ports": [PORT]},
        ),
    },
    '/v1/ports/?fields=uuid,address':
    {
        'GET': (
            {},
            {"ports": [PORT]},
        ),
    },
    '/v1/ports/%s' % PORT['uuid']:
    {
        'GET': (
            {},
            PORT,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_PORT,
        ),
    },
    '/v1/ports/%s?fields=uuid,address' % PORT['uuid']:
    {
        'GET': (
            {},
            PORT,
        ),
    },
    '/v1/ports/detail?address=%s' % PORT['address']:
    {
        'GET': (
            {},
            {"ports": [PORT]},
        ),
    },
    '/v1/ports/?address=%s' % PORT['address']:
    {
        'GET': (
            {},
            {"ports": [PORT]},
        ),
    },
    '/v1/ports/?node=%s' % PORT['node_uuid']:
    {
        'GET': (
            {},
            {"ports": [PORT]},
        ),
    },
    '/v1/ports/?portgroup=%s' % PORT['portgroup_uuid']:
    {
        'GET': (
            {},
            {"ports": [PORT]},
        ),
    },
}

fake_responses_pagination = {
    '/v1/ports':
    {
        'GET': (
            {},
            {"ports": [PORT],
             "next": "http://127.0.0.1:6385/v1/ports/?limit=1"}
        ),
    },
    '/v1/ports/?limit=1':
    {
        'GET': (
            {},
            {"ports": [PORT2]}
        ),
    },
    '/v1/ports/?marker=%s' % PORT['uuid']:
    {
        'GET': (
            {},
            {"ports": [PORT2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/ports/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"ports": [PORT2, PORT]}
        ),
    },
    '/v1/ports/?sort_dir=desc':
    {
        'GET': (
            {},
            {"ports": [PORT2, PORT]}
        ),
    },
}


class PortManagerTest(testtools.TestCase):

    def setUp(self) -> None:
        super(PortManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = ironicclient.v1.port.PortManager(self.api)

    def test_ports_list(self) -> None:
        ports = self.mgr.list()
        expect = [
            ('GET', '/v1/ports', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(ports))

    def test_ports_list_by_address(self) -> None:
        ports = self.mgr.list(address=PORT['address'])
        expect = [
            ('GET', '/v1/ports/?address=%s' % PORT['address'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(ports))

    def test_ports_list_by_address_detail(self) -> None:
        ports = self.mgr.list(address=PORT['address'], detail=True)
        expect = [
            ('GET', '/v1/ports/detail?address=%s' % PORT['address'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(ports))

    def test_ports_list_by_node(self) -> None:
        ports = self.mgr.list(node=PORT['node_uuid'])
        expect = [
            ('GET', '/v1/ports/?node=%s' % PORT['node_uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(ports))

    def test_ports_list_by_portgroup(self) -> None:
        ports = self.mgr.list(portgroup=PORT['portgroup_uuid'])
        expect = [
            ('GET', '/v1/ports/?portgroup=%s' % PORT['portgroup_uuid'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(ports))

    def test_ports_list_detail(self) -> None:
        ports = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/ports/detail', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(ports))

    def test_port_list_fields(self) -> None:
        ports = self.mgr.list(fields=['uuid', 'address'])
        expect = [
            ('GET', '/v1/ports/?fields=uuid,address', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(ports))

    def test_port_list_detail_and_fields_fail(self) -> None:
        self.assertRaises(exc.InvalidAttribute, self.mgr.list,
                          detail=True, fields=['uuid', 'address'])

    def test_ports_list_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.port.PortManager(self.api)
        ports = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/ports/?limit=1', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(ports, HasLength(1))

    def test_ports_list_marker(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.port.PortManager(self.api)
        ports = self.mgr.list(marker=PORT['uuid'])
        expect = [
            ('GET', '/v1/ports/?marker=%s' % PORT['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(ports, HasLength(1))

    def test_ports_list_pagination_no_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.port.PortManager(self.api)
        ports = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/ports', {}, None),
            ('GET', '/v1/ports/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(ports, HasLength(2))

    def test_ports_list_sort_key(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.port.PortManager(self.api)
        ports = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/ports/?sort_key=updated_at', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(ports))

    def test_ports_list_sort_dir(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.port.PortManager(self.api)
        ports = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/ports/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(ports))

    def test_ports_show(self) -> None:
        port = self.mgr.get(PORT['uuid'])
        expect = [
            ('GET', '/v1/ports/%s' % PORT['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(PORT['uuid'], port.uuid)
        self.assertEqual(PORT['address'], port.address)
        self.assertEqual(PORT['node_uuid'], port.node_uuid)
        self.assertEqual(PORT['pxe_enabled'], port.pxe_enabled)
        self.assertEqual(PORT['local_link_connection'],
                         port.local_link_connection)
        self.assertEqual(PORT['portgroup_uuid'], port.portgroup_uuid)
        self.assertEqual(PORT['physical_network'], port.physical_network)
        self.assertEqual(PORT['is_smartnic'], port.is_smartnic)
        self.assertEqual(PORT['name'], port.name)

    def test_ports_show_by_address(self) -> None:
        port = self.mgr.get_by_address(PORT['address'])
        expect = [
            ('GET', '/v1/ports/detail?address=%s' % PORT['address'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(PORT['uuid'], port.uuid)
        self.assertEqual(PORT['address'], port.address)
        self.assertEqual(PORT['node_uuid'], port.node_uuid)
        self.assertEqual(PORT['pxe_enabled'], port.pxe_enabled)
        self.assertEqual(PORT['local_link_connection'],
                         port.local_link_connection)
        self.assertEqual(PORT['portgroup_uuid'], port.portgroup_uuid)
        self.assertEqual(PORT['physical_network'], port.physical_network)
        self.assertEqual(PORT['is_smartnic'], port.is_smartnic)

    def test_port_show_fields(self) -> None:
        port = self.mgr.get(PORT['uuid'], fields=['uuid', 'address'])
        expect = [
            ('GET', '/v1/ports/%s?fields=uuid,address' %
             PORT['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(PORT['uuid'], port.uuid)
        self.assertEqual(PORT['address'], port.address)

    def test_create(self) -> None:
        port = self.mgr.create(**CREATE_PORT)
        expect = [
            ('POST', '/v1/ports', {}, CREATE_PORT),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(port)

    def test_create_with_uuid(self) -> None:
        port = self.mgr.create(**CREATE_PORT_WITH_UUID)
        expect = [
            ('POST', '/v1/ports', {}, CREATE_PORT_WITH_UUID),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(port)

    def test_delete(self) -> None:
        self.mgr.delete(port_id=PORT['uuid'])
        expect = [
            ('DELETE', '/v1/ports/%s' % PORT['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)

    def test_update(self) -> None:
        patch = {'op': 'replace',
                 'value': NEW_ADDR,
                 'path': '/address'}
        port = self.mgr.update(port_id=PORT['uuid'], patch=patch)
        expect = [
            ('PATCH', '/v1/ports/%s' % PORT['uuid'], {}, patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_ADDR, port.address)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import copy
from unittest import mock

import testtools

from ironicclient import exc
from ironicclient.tests.unit import utils
import ironicclient.v1.allocation

ALLOCATION = {'uuid': '11111111-2222-3333-4444-555555555555',
              'name': 'Allocation-name',
              'owner': None,
              'state': 'active',
              'node_uuid': '66666666-7777-8888-9999-000000000000',
              'last_error': None,
              'resource_class': 'baremetal',
              'traits': [],
              'candidate_nodes': [],
              'extra': {}}

ALLOCATION2 = {'uuid': '55555555-4444-3333-2222-111111111111',
               'name': 'Allocation2-name',
               'owner': 'fake-owner',
               'state': 'allocating',
               'node_uuid': None,
               'last_error': None,
               'resource_class': 'baremetal',
               'traits': [],
               'candidate_nodes': [],
               'extra': {}}

CREATE_ALLOCATION = copy.deepcopy(ALLOCATION)
for field in ('state', 'node_uuid', 'last_error'):
    del CREATE_ALLOCATION[field]

fake_responses = {
    '/v1/allocations':
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION, ALLOCATION2]},
        ),
        'POST': (
            {},
            CREATE_ALLOCATION,
        ),
    },
    '/v1/allocations/%s' % ALLOCATION['uuid']:
    {
        'GET': (
            {},
            ALLOCATION,
        ),
        'DELETE': (
            {},
            None,
        ),
    },
    '/v1/allocations/?node=%s' % ALLOCATION['node_uuid']:
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION]},
        ),
    },
    '/v1/allocations/?owner=%s' % ALLOCATION2['owner']:
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION2]},
        ),
    },
}

fake_responses_pagination = {
    '/v1/allocations':
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION],
             "next": "http://127.0.0.1:6385/v1/allocations/?limit=1"}
        ),
    },
    '/v1/allocations/?limit=1':
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION2]}
        ),
    },
    '/v1/allocations/?marker=%s' % ALLOCATION['uuid']:
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/allocations/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION2, ALLOCATION]}
        ),
    },
    '/v1/allocations/?sort_dir=desc':
    {
        'GET': (
            {},
            {"allocations": [ALLOCATION2, ALLOCATION]}
        ),
    },
}


class AllocationManagerTest(testtools.TestCase):

    def setUp(self) -> None:
        super(AllocationManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = ironicclient.v1.allocation.AllocationManager(self.api)

    def test_allocations_list(self) -> None:
        allocations = self.mgr.list()
        expect = [
            ('GET', '/v1/allocations', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(allocations))

        expected_resp = ({}, {"allocations": [ALLOCATION, ALLOCATION2]},)
        self.assertEqual(expected_resp,
                         self.api.responses['/v1/allocations']['GET'])

    def test_allocations_list_by_node(self) -> None:
        allocations = self.mgr.list(node=ALLOCATION['node_uuid'])
        expect = [
            ('GET', '/v1/allocations/?node=%s' % ALLOCATION['node_uuid'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(allocations))

        expected_resp = ({}, {"allocations": [ALLOCATION, ALLOCATION2]},)
        self.assertEqual(expected_resp,
                         self.api.responses['/v1/allocations']['GET'])

    def test_allocations_list_by_owner(self) -> None:
        allocations = self.mgr.list(owner=ALLOCATION2['owner'])
        expect = [
            ('GET', '/v1/allocations/?owner=%s' % ALLOCATION2['owner'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(allocations))

        expected_resp = ({}, {"allocations": [ALLOCATION, ALLOCATION2]},)
        self.assertEqual(expected_resp,
                         self.api.responses['/v1/allocations']['GET'])

    def test_allocations_show(self) -> None:
        allocation = self.mgr.get(ALLOCATION['uuid'])
        expect = [
            ('GET', '/v1/allocations/%s' % ALLOCATION['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(ALLOCATION['uuid'], allocation.uuid)
        self.assertEqual(ALLOCATION['name'], allocation.name)
        self.assertEqual(ALLOCATION['owner'], allocation.owner)
        self.assertEqual(ALLOCATION['node_uuid'], allocation.node_uuid)
        self.assertEqual(ALLOCATION['state'], allocation.state)
        self.assertEqual(ALLOCATION['resource_class'],
                         allocation.resource_class)

        expected_resp = ({}, ALLOCATION,)
        self.assertEqual(
            expected_resp,
            self.api.responses['/v1/allocations/%s'
                               % ALLOCATION['uuid']]['GET'])

    def test_create(self) -> None:
        allocation = self.mgr.create(**CREATE_ALLOCATION)
        expect = [
            ('POST', '/v1/allocations', {}, CREATE_ALLOCATION),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(allocation)

        self.assertIn(
            ALLOCATION,
            self.api.responses['/v1/allocations']['GET'][1]['allocations'])

    def test_delete(self) -> None:
        self.mgr.delete(allocation_id=ALLOCATION['uuid'])
        expect = [
            ('DELETE', '/v1/allocations/%s' % ALLOCATION['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)

        expected_resp = ({}, ALLOCATION,)
        self.assertEqual(
            expected_resp,
            self.api.responses['/v1/allocations/%s'
                               % ALLOCATION['uuid']]['GET'])


class AllocationManagerPaginationTest(testtools.TestCase):

    def setUp(self) -> None:
        super(AllocationManagerPaginationTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.allocation.AllocationManager(self.api)

    def test_allocations_list_limit(self) -> None:
        allocations = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/allocations/?limit=1', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(allocations))

        expected_resp = (
            {}, {"next": "http://127.0.0.1:6385/v1/allocations/?limit=1",
                 "allocations": [ALLOCATION]},)
        self.assertEqual(expected_resp,
                         self.api.responses['/v1/allocations']['GET'])

    def test_allocations_list_marker(self) -> None:
        allocations = self.mgr.list(marker=ALLOCATION['uuid'])
        expect = [
            ('GET', '/v1/allocations/?marker=%s' % ALLOCATION['uuid'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(allocations))

        expected_resp = (
            {}, {"next": "http://127.0.0.1:6385/v1/allocations/?limit=1",
                 "allocations": [ALLOCATION]},)
        self.assertEqual(expected_resp,
                         self.api.responses['/v1/allocations']['GET'])

    def test_allocations_list_pagination_no_limit(self) -> None:
        allocations = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/allocations', {}, None),
            ('GET', '/v1/allocations/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(allocations))

        expected_resp = (
            {}, {"next": "http://127.0.0.1:6385/v1/allocations/?limit=1",
                 "allocations": [ALLOCATION]},)
        self.assertEqual(expected_resp,
                         self.api.responses['/v1/allocations']['GET'])


class AllocationManagerSortingTest(testtools.TestCase):

    def setUp(self) -> None:
        super(AllocationManagerSortingTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.allocation.AllocationManager(self.api)

    def test_allocations_list_sort_key(self) -> None:
        allocations = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/allocations/?sort_key=updated_at', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(allocations))

        expected_resp = ({}, {"allocations": [ALLOCATION2, ALLOCATION]},)
        self.assertEqual(
            expected_resp,
            self.api.responses['/v1/allocations/?sort_key=updated_at']['GET'])

    def test_allocations_list_sort_dir(self) -> None:
        allocations = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/allocations/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(allocations))

        expected_resp = ({}, {"allocations": [ALLOCATION2, ALLOCATION]},)
        self.assertEqual(
            expected_resp,
            self.api.responses['/v1/allocations/?sort_dir=desc']['GET'])


@mock.patch('time.sleep', autospec=True)
@mock.patch('ironicclient.v1.allocation.AllocationManager.get', autospec=True)
class AllocationWaitTest(testtools.TestCase):

    def setUp(self) -> None:
        super(AllocationWaitTest, self).setUp()
        self.mgr = ironicclient.v1.allocation.AllocationManager(mock.Mock())

    def _fake_allocation(
        self, state: str, error: str | None = None,
    ) -> mock.Mock:
        return mock.Mock(state=state, last_error=error)

    def test_success(
        self, mock_get: mock.MagicMock, mock_sleep: mock.MagicMock,
    ) -> None:
        allocations = [
            self._fake_allocation('allocating'),
            self._fake_allocation('allocating'),
            self._fake_allocation('active'),
        ]
        mock_get.side_effect = allocations

        result = self.mgr.wait('alloc1')
        self.assertIs(result, allocations[2])
        self.assertEqual(3, mock_get.call_count)
        self.assertEqual(2, mock_sleep.call_count)
        mock_get.assert_called_with(
            self.mgr, 'alloc1', os_ironic_api_version=None,
            global_request_id=None)

    def test_error(
        self, mock_get: mock.MagicMock, mock_sleep: mock.MagicMock,
    ) -> None:
        allocations = [
            self._fake_allocation('allocating'),
            self._fake_allocation('error'),
        ]
        mock_get.side_effect = allocations

        self.assertRaises(exc.StateTransitionFailed,
                          self.mgr.wait, 'alloc1')

        self.assertEqual(2, mock_get.call_count)
        self.assertEqual(1, mock_sleep.call_count)
        mock_get.assert_called_with(
            self.mgr, 'alloc1', os_ironic_api_version=None,
            global_request_id=None)

    def test_timeout(
        self, mock_get: mock.MagicMock, mock_sleep: mock.MagicMock,
    ) -> None:
        mock_get.return_value = self._fake_allocation('allocating')

        self.assertRaises(exc.StateTransitionTimeout,
                          self.mgr.wait, 'alloc1', timeout=0.001)

        mock_get.assert_called_with(
            self.mgr, 'alloc1', os_ironic_api_version=None,
            global_request_id=None)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import copy
from unittest import mock

from osc_lib.tests import utils as osctestutils

from ironicclient import exc
from ironicclient.osc.v1 import baremetal_volume_connector as bm_vol_connector
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestBaremetalVolumeConnector(baremetal_fakes.TestBaremetal):

    def setUp(self) -> None:
        super(TestBaremetalVolumeConnector, self).setUp()

        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()


class TestCreateBaremetalVolumeConnector(TestBaremetalVolumeConnector):

    def setUp(self) -> None:
        super(TestCreateBaremetalVolumeConnector, self).setUp()

        self.baremetal_mock.volume_connector.create.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.VOLUME_CONNECTOR),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = (
            bm_vol_connector.CreateBaremetalVolumeConnector(self.app, None))

    def test_baremetal_volume_connector_create(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
            '--type', baremetal_fakes.baremetal_volume_connector_type,
            '--connector-id',
            baremetal_fakes.baremetal_volume_connector_connector_id,
            '--uuid', baremetal_fakes.baremetal_volume_connector_uuid,
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('type', baremetal_fakes.baremetal_volume_connector_type),
            ('connector_id',
             baremetal_fakes.baremetal_volume_connector_connector_id),
            ('uuid', baremetal_fakes.baremetal_volume_connector_uuid),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'type': baremetal_fakes.baremetal_volume_connector_type,
            'connector_id':
                baremetal_fakes.baremetal_volume_connector_connector_id,
            'uuid': baremetal_fakes.baremetal_volume_connector_uuid,
        }

        self.baremetal_mock.volume_connector.create.assert_called_once_with(
            **args)

    def test_baremetal_volume_connector_create_without_uuid(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
            '--type', baremetal_fakes.baremetal_volume_connector_type,
            '--connector-id',
            baremetal_fakes.baremetal_volume_connector_connector_id,
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('type', baremetal_fakes.baremetal_volume_connector_type),
            ('connector_id',
             baremetal_fakes.baremetal_volume_connector_connector_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'type': baremetal_fakes.baremetal_volume_connector_type,
            'connector_id':
                baremetal_fakes.baremetal_volume_connector_connector_id,
        }

        self.baremetal_mock.volume_connector.create.assert_called_once_with(
            **args)

    def test_baremetal_volume_connector_create_extras(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
            '--type', baremetal_fakes.baremetal_volume_connector_type,
            '--connector-id',
            baremetal_fakes.baremetal_volume_connector_connector_id,
            '--extra', 'key1=value1',
            '--extra', 'key2=value2',
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('type', baremetal_fakes.baremetal_volume_connector_type),
            ('connector_id',
             baremetal_fakes.baremetal_volume_connector_connector_id),
            ('extra', ['key1=value1', 'key2=value2'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'type': baremetal_fakes.baremetal_volume_connector_type,
            'connector_id':
                baremetal_fakes.baremetal_volume_connector_connector_id,
            'extra': baremetal_fakes.baremetal_volume_connector_extra,
        }

        self.baremetal_mock.volume_connector.create.assert_called_once_with(
            **args)

    def test_baremetal_volume_connector_create_invalid_type(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
            '--type', 'invalid',
            '--connector-id',
            baremetal_fakes.baremetal_volume_connector_connector_id,
            '--uuid', baremetal_fakes.baremetal_volume_connector_uuid,
        ]
        verifylist = None

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_create_missing_node(self) -> None:
        arglist = [
            '--type', baremetal_fakes.baremetal_volume_connector_type,
            '--connector-id',
            baremetal_fakes.baremetal_volume_connector_connector_id,
            '--uuid', baremetal_fakes.baremetal_volume_connector_uuid,
        ]
        verifylist = None

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_create_missing_type(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
            '--connector-id',
            baremetal_fakes.baremetal_volume_connector_connector_id,
            '--uuid', baremetal_fakes.baremetal_volume_connector_uuid,
        ]
        verifylist = None

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_create_missing_connector_id(
            self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
            '--type', baremetal_fakes.baremetal_volume_connector_type,
            '--uuid', baremetal_fakes.baremetal_volume_connector_uuid,
        ]
        verifylist = None

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestShowBaremetalVolumeConnector(TestBaremetalVolumeConnector):

    def setUp(self) -> None:
        super(TestShowBaremetalVolumeConnector, self).setUp()

        self.baremetal_mock.volume_connector.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.VOLUME_CONNECTOR),
                loaded=True))

        self.cmd = (
            bm_vol_connector.ShowBaremetalVolumeConnector(self.app, None))

    def test_baremetal_volume_connector_show(self) -> None:
        arglist = ['vvv-cccccc-vvvv']
        verifylist = [('volume_connector',
                       baremetal_fakes.baremetal_volume_connector_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        args = ['vvv-cccccc-vvvv']
        self.baremetal_mock.volume_connector.get.assert_called_once_with(
            *args, fields=None)
        collist = ('connector_id', 'extra', 'node_uuid', 'type', 'uuid')
        self.assertEqual(collist, columns)

        datalist = (
            baremetal_fakes.baremetal_volume_connector_connector_id,
            baremetal_fakes.baremetal_volume_connector_extra,
            baremetal_fakes.baremetal_uuid,
            baremetal_fakes.baremetal_volume_connector_type,
            baremetal_fakes.baremetal_volume_connector_uuid,
        )
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_volume_connector_show_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_show_fields(self) -> None:
        arglist = ['vvv-cccccc-vvvv', '--fields', 'uuid', 'connector_id']
        verifylist = [('fields', [['uuid', 'connector_id']]),
                      ('volume_connector',
                       baremetal_fakes.baremetal_volume_connector_uuid)]

        fake_vc = copy.deepcopy(baremetal_fakes.VOLUME_CONNECTOR)
        fake_vc.pop('type')
        fake_vc.pop('extra')
        fake_vc.pop('node_uuid')
        self.baremetal_mock.volume_connector.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                fake_vc,
                loaded=True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        args = ['vvv-cccccc-vvvv']
        fields = ['uuid', 'connector_id']
        self.baremetal_mock.volume_connector.get.assert_called_once_with(
            *args, fields=fields)
        collist = ('connector_id', 'uuid')
        self.assertEqual(collist, columns)

        datalist = (
            baremetal_fakes.baremetal_volume_connector_connector_id,
            baremetal_fakes.baremetal_volume_connector_uuid,
        )
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_volume_connector_show_fields_multiple(self) -> None:
        arglist = ['vvv-cccccc-vvvv', '--fields', 'uuid', 'connector_id',
                   '--fields', 'type']
        verifylist = [('fields', [['uuid', 'connector_id'], ['type']]),
                      ('volume_connector',
                       baremetal_fakes.baremetal_volume_connector_uuid)]

        fake_vc = copy.deepcopy(baremetal_fakes.VOLUME_CONNECTOR)
        fake_vc.pop('extra')
        fake_vc.pop('node_uuid')
        self.baremetal_mock.volume_connector.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                fake_vc,
                loaded=True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        args = ['vvv-cccccc-vvvv']
        fields = ['uuid', 'connector_id', 'type']
        self.baremetal_mock.volume_connector.get.assert_called_once_with(
            *args, fields=fields)
        collist = ('connector_id', 'type', 'uuid')
        self.assertEqual(collist, columns)

        datalist = (
            baremetal_fakes.baremetal_volume_connector_connector_id,
            baremetal_fakes.baremetal_volume_connector_type,
            baremetal_fakes.baremetal_volume_connector_uuid,
        )
        self.assertEqual(datalist, tuple(data))


class TestListBaremetalVolumeConnector(TestBaremetalVolumeConnector):
    def setUp(self) -> None:
        super(TestListBaremetalVolumeConnector, self).setUp()

        self.baremetal_mock.volume_connector.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.VOLUME_CONNECTOR),
                loaded=True)
        ]
        self.cmd = (
            bm_vol_connector.ListBaremetalVolumeConnector(self.app, None))

    def test_baremetal_volume_connector_list(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None}
        self.baremetal_mock.volume_connector.list.assert_called_once_with(
            **kwargs)

        collist = (
            "uuid",
            "node_uuid",
            "type",
            "connector_id")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_volume_connector_uuid,
                     baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_volume_connector_type,
                     baremetal_fakes.baremetal_volume_connector_connector_id),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_volume_connector_list_node(self) -> None:
        arglist = ['--node', baremetal_fakes.baremetal_uuid]
        verifylist = [('node', baremetal_fakes.baremetal_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'node': baremetal_fakes.baremetal_uuid,
            'marker': None,
            'limit': None}
        self.baremetal_mock.volume_connector.list.assert_called_once_with(
            **kwargs)

        collist = (
            "uuid",
            "node_uuid",
            "type",
            "connector_id")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_volume_connector_uuid,
                     baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_volume_connector_type,
                     baremetal_fakes.baremetal_volume_connector_connector_id),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_volume_connector_list_long(self) -> None:
        arglist = ['--long']
        verifylist = [('detail', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'detail': True,
            'marker': None,
            'limit': None,
        }
        self.baremetal_mock.volume_connector.list.assert_called_with(**kwargs)

        collist = ('uuid', 'node_uuid', 'type', 'connector_id', 'extra',
                   'created_at', 'updated_at')
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_volume_connector_uuid,
                     baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_volume_connector_type,
                     baremetal_fakes.baremetal_volume_connector_connector_id,
                     baremetal_fakes.baremetal_volume_connector_extra,
                     '',
                     ''),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_volume_connector_list_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'connector_id']
        verifylist = [('fields', [['uuid', 'connector_id']])]

        fake_vc = copy.deepcopy(baremetal_fakes.VOLUME_CONNECTOR)
        fake_vc.pop('type')
        fake_vc.pop('extra')
        fake_vc.pop('node_uuid')
        self.baremetal_mock.volume_connector.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                fake_vc,
                loaded=True)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'detail': False,
            'marker': None,
            'limit': None,
            'fields': ('uuid', 'connector_id')
        }
        self.baremetal_mock.volume_connector.list.assert_called_with(**kwargs)

        collist = ('uuid', 'connector_id')
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_volume_connector_uuid,
                     baremetal_fakes.baremetal_volume_connector_connector_id),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_volume_connector_list_fields_multiple(self) -> None:
        arglist = ['--fields', 'uuid', 'connector_id', '--fields', 'extra']
        verifylist = [('fields', [['uuid', 'connector_id'], ['extra']])]

        fake_vc = copy.deepcopy(baremetal_fakes.VOLUME_CONNECTOR)
        fake_vc.pop('type')
        fake_vc.pop('node_uuid')
        self.baremetal_mock.volume_connector.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                fake_vc,
                loaded=True)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'detail': False,
            'marker': None,
            'limit': None,
            'fields': ('uuid', 'connector_id', 'extra')
        }
        self.baremetal_mock.volume_connector.list.assert_called_with(**kwargs)

        collist = ('uuid', 'connector_id', 'extra')
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_volume_connector_uuid,
                     baremetal_fakes.baremetal_volume_connector_connector_id,
                     baremetal_fakes.baremetal_volume_connector_extra),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_volume_connector_list_invalid_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'invalid']
        verifylist = [('fields', [['uuid', 'invalid']])]
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_list_marker(self) -> None:
        arglist = ['--marker', baremetal_fakes.baremetal_volume_connector_uuid]
        verifylist = [
            ('marker', baremetal_fakes.baremetal_volume_connector_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': baremetal_fakes.baremetal_volume_connector_uuid,
            'limit': None}
        self.baremetal_mock.volume_connector.list.assert_called_once_with(
            **kwargs)

    def test_baremetal_volume_connector_list_limit(self) -> None:
        arglist = ['--limit', '10']
        verifylist = [('limit', 10)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': 10}
        self.baremetal_mock.volume_connector.list.assert_called_once_with(
            **kwargs)

    def test_baremetal_volume_connector_list_sort(self) -> None:
        arglist = ['--sort', 'type']
        verifylist = [('sort', 'type')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None}
        self.baremetal_mock.volume_connector.list.assert_called_once_with(
            **kwargs)

    def test_baremetal_volume_connector_list_sort_desc(self) -> None:
        arglist = ['--sort', 'type:desc']
        verifylist = [('sort', 'type:desc')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None}
        self.baremetal_mock.volume_connector.list.assert_called_once_with(
            **kwargs)

    def test_baremetal_volume_connector_list_exclusive_options(self) -> None:
        arglist = ['--fields', 'uuid', '--long']
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, [])

    def test_baremetal_volume_connector_list_negative_limit(self) -> None:
        arglist = ['--limit', '-1']
        verifylist = [('limit', -1)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exc.CommandError,
                          self.cmd.take_action,
                          parsed_args)


class TestDeleteBaremetalVolumeConnector(TestBaremetalVolumeConnector):

    def setUp(self) -> None:
        super(TestDeleteBaremetalVolumeConnector, self).setUp()

        self.cmd = (
            bm_vol_connector.DeleteBaremetalVolumeConnector(self.app, None))

    def test_baremetal_volume_connector_delete(self) -> None:
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid]
        verifylist = [('volume_connectors',
                       [baremetal_fakes.baremetal_volume_connector_uuid])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.baremetal_mock.volume_connector.delete.assert_called_with(
            baremetal_fakes.baremetal_volume_connector_uuid)

    def test_baremetal_volume_connector_delete_multiple(self) -> None:
        fake_volume_connector_uuid2 = 'vvv-cccccc-cccc'
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid,
                   fake_volume_connector_uuid2]
        verifylist = [('volume_connectors',
                       [baremetal_fakes.baremetal_volume_connector_uuid,
                        fake_volume_connector_uuid2])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.baremetal_mock.volume_connector.delete.assert_has_calls(
            [mock.call(baremetal_fakes.baremetal_volume_connector_uuid),
             mock.call(fake_volume_connector_uuid2)])
        self.assertEqual(
            2, self.baremetal_mock.volume_connector.delete.call_count)

    def test_baremetal_volume_connector_delete_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_delete_error(self) -> None:
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid]
        verifylist = [('volume_connectors',
                       [baremetal_fakes.baremetal_volume_connector_uuid])]

        self.baremetal_mock.volume_connector.delete.side_effect = (
            exc.NotFound())

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exc.ClientException,
                          self.cmd.take_action,
                          parsed_args)
        self.baremetal_mock.volume_connector.delete.assert_called_with(
            baremetal_fakes.baremetal_volume_connector_uuid)

    def test_baremetal_volume_connector_delete_multiple_error(self) -> None:
        fake_volume_connector_uuid2 = 'vvv-cccccc-cccc'
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid,
                   fake_volume_connector_uuid2]
        verifylist = [('volume_connectors',
                       [baremetal_fakes.baremetal_volume_connector_uuid,
                        fake_volume_connector_uuid2])]

        self.baremetal_mock.volume_connector.delete.side_effect = [
            None, exc.NotFound()]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exc.ClientException,
                          self.cmd.take_action,
                          parsed_args)

        self.baremetal_mock.volume_connector.delete.assert_has_calls(
            [mock.call(baremetal_fakes.baremetal_volume_connector_uuid),
             mock.call(fake_volume_connector_uuid2)])
        self.assertEqual(
            2, self.baremetal_mock.volume_connector.delete.call_count)


class TestSetBaremetalVolumeConnector(TestBaremetalVolumeConnector):
    def setUp(self) -> None:
        super(TestSetBaremetalVolumeConnector, self).setUp()

        self.cmd = (
            bm_vol_connector.SetBaremetalVolumeConnector(self.app, None))

    def test_baremetal_volume_connector_set_node_uuid(self) -> None:
        new_node_uuid = 'xxx-xxxxxx-zzzz'
        arglist = [
            baremetal_fakes.baremetal_volume_connector_uuid,
            '--node', new_node_uuid]
        verifylist = [
            ('volume_connector',
             baremetal_fakes.baremetal_volume_connector_uuid),
            ('node_uuid', new_node_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/node_uuid', 'value': new_node_uuid, 'op': 'add'}])

    def test_baremetal_volume_connector_set_type(self) -> None:
        new_type = 'wwnn'
        arglist = [
            baremetal_fakes.baremetal_volume_connector_uuid,
            '--type', new_type]
        verifylist = [
            ('volume_connector',
             baremetal_fakes.baremetal_volume_connector_uuid),
            ('type', new_type)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/type', 'value': new_type, 'op': 'add'}])

    def test_baremetal_volume_connector_set_invalid_type(self) -> None:
        new_type = 'invalid'
        arglist = [
            baremetal_fakes.baremetal_volume_connector_uuid,
            '--type', new_type]
        verifylist = None

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_set_connector_id(self) -> None:
        new_conn_id = '11:22:33:44:55:66:77:88'
        arglist = [
            baremetal_fakes.baremetal_volume_connector_uuid,
            '--connector-id', new_conn_id]
        verifylist = [
            ('volume_connector',
             baremetal_fakes.baremetal_volume_connector_uuid),
            ('connector_id', new_conn_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/connector_id', 'value': new_conn_id, 'op': 'add'}])

    def test_baremetal_volume_connector_set_type_and_connector_id(
            self) -> None:
        new_type = 'wwnn'
        new_conn_id = '11:22:33:44:55:66:77:88'
        arglist = [
            baremetal_fakes.baremetal_volume_connector_uuid,
            '--type', new_type,
            '--connector-id', new_conn_id]
        verifylist = [
            ('volume_connector',
             baremetal_fakes.baremetal_volume_connector_uuid),
            ('type', new_type),
            ('connector_id', new_conn_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/type', 'value': new_type, 'op': 'add'},
             {'path': '/connector_id', 'value': new_conn_id, 'op': 'add'}])

    def test_baremetal_volume_connector_set_extra(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_volume_connector_uuid,
            '--extra', 'foo=bar']
        verifylist = [
            ('volume_connector',
             baremetal_fakes.baremetal_volume_connector_uuid),
            ('extra', ['foo=bar'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/extra/foo', 'value': 'bar', 'op': 'add'}])

    def test_baremetal_volume_connector_set_multiple_extras(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_volume_connector_uuid,
            '--extra', 'key1=val1', '--extra', 'key2=val2']
        verifylist = [
            ('volume_connector',
             baremetal_fakes.baremetal_volume_connector_uuid),
            ('extra', ['key1=val1', 'key2=val2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/extra/key1', 'value': 'val1', 'op': 'add'},
             {'path': '/extra/key2', 'value': 'val2', 'op': 'add'}])

    def test_baremetal_volume_connector_set_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_set_no_property(self) -> None:
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid]
        verifylist = [('volume_connector',
                       baremetal_fakes.baremetal_volume_connector_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_not_called()


class TestUnsetBaremetalVolumeConnector(TestBaremetalVolumeConnector):
    def setUp(self) -> None:
        super(TestUnsetBaremetalVolumeConnector, self).setUp()

        self.cmd = (
            bm_vol_connector.UnsetBaremetalVolumeConnector(self.app, None))

    def test_baremetal_volume_connector_unset_extra(self) -> None:
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid,
                   '--extra', 'key1']
        verifylist = [('volume_connector',
                       baremetal_fakes.baremetal_volume_connector_uuid),
                      ('extra', ['key1'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/extra/key1', 'op': 'remove'}])

    def test_baremetal_volume_connector_unset_multiple_extras(self) -> None:
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid,
                   '--extra', 'key1', '--extra', 'key2']
        verifylist = [('volume_connector',
                       baremetal_fakes.baremetal_volume_connector_uuid),
                      ('extra', ['key1', 'key2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_called_once_with(
            baremetal_fakes.baremetal_volume_connector_uuid,
            [{'path': '/extra/key1', 'op': 'remove'},
             {'path': '/extra/key2', 'op': 'remove'}])

    def test_baremetal_volume_connector_unset_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_volume_connector_unset_no_property(self) -> None:
        arglist = [baremetal_fakes.baremetal_volume_connector_uuid]
        verifylist = [('volume_connector',
                       baremetal_fakes.baremetal_volume_connector_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.baremetal_mock.volume_connector.update.assert_not_called()



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Typed base command classes for ironicclient OSC commands.

These subclass osc-lib's command classes and annotate ``app`` so that
``self.app.client_manager.baremetal`` is statically known to be
:class:`ironicclient.v1.client.Client`, removing the need for
``# type: ignore[attr-defined]`` throughout the command implementations.

See https://review.opendev.org/c/openstack/python-openstackclient/+/970681
for the equivalent pattern in python-openstackclient.
"""

from __future__ import annotations

from typing import Any

from cliff import lister
from cliff import show
from osc_lib import clientmanager
from osc_lib.command import command
from osc_lib import shell


class _ClientManager(clientmanager.ClientManager):
    # NOTE(anandkaranubc): baremetal is injected dynamically by the plugin
    # system (via setattr in osc-lib's get_plugin_modules). Typed as
    # Any because osc-lib's utility functions (e.g. get_item_properties)
    # have narrow type signatures that don't accept Resource objects.
    # Can be narrowed to v1.client.Client once those callers are
    # updated.
    baremetal: Any


class _App(shell.OpenStackShell):
    client_manager: _ClientManager


class Command(command.Command):
    app: _App


class Lister(Command, lister.Lister):
    pass


class ShowOne(Command, show.ShowOne):
    pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import unittest
from unittest import mock
from unittest.mock import patch

from ironicclient.v1.shard import Shard
from ironicclient.v1.shard import ShardManager


class TestShardManager(unittest.TestCase):

    @patch('ironicclient.common.base.Manager._list')
    def test_list_shards(self, mock_list: mock.MagicMock) -> None:
        # Initialize the ShardManager
        shard_manager = ShardManager(api=None)  # `api=None` for simplicity

        # Mock response for the list of shards
        mock_response = [
            Shard(shard_manager, {'name': 'example_shard1', 'count': 47}),
            Shard(shard_manager, {'name': 'example_shard2', 'count': 46}),
            Shard(shard_manager, {'name': None, 'count': 3}),
            # Nodes with no shard assigned
        ]

        # Configure mock to return the mocked response
        mock_list.return_value = mock_response

        # Perform the test call
        result = shard_manager.list(os_ironic_api_version="1.82")

        # Assertions
        mock_list.assert_called_once_with(
            shard_manager._path(None),
            "shards",
            os_ironic_api_version="1.82",
            global_request_id=None
        )
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].name, 'example_shard1')
        self.assertEqual(result[0].count, 47)
        self.assertEqual(result[1].name, 'example_shard2')
        self.assertEqual(result[1].count, 46)
        self.assertIsNone(result[2].name)
        self.assertEqual(result[2].count, 3)

    @patch('ironicclient.common.base.Manager._list')
    def test_list_shards_empty(self, mock_list: mock.MagicMock) -> None:
        # Test when the shards list is empty
        mock_list.return_value = []

        # Initialize the ShardManager
        shard_manager = ShardManager(api=None)

        # Perform the test call
        result = shard_manager.list(os_ironic_api_version="1.82")

        # Assertions
        mock_list.assert_called_once_with(
            shard_manager._path(None),
            "shards",
            os_ironic_api_version="1.82",
            global_request_id=None
        )
        self.assertEqual(result, [])

    @patch('ironicclient.common.base.Manager._list')
    def test_list_shards_with_global_request_id(
        self, mock_list: mock.MagicMock,
    ) -> None:
        # Initialize the ShardManager
        shard_manager = ShardManager(api=None)

        # Test with global request ID
        mock_response = [
            Shard(shard_manager, {'name': 'example_shard1', 'count': 47}),
            Shard(shard_manager, {'name': 'example_shard2', 'count': 46}),
        ]
        mock_list.return_value = mock_response

        # Perform the test call with global_request_id
        result = shard_manager.list(
            os_ironic_api_version="1.82",
            global_request_id="req-12345"
        )

        # Assertions
        mock_list.assert_called_once_with(
            shard_manager._path(None),
            "shards",
            os_ironic_api_version="1.82",
            global_request_id="req-12345"
        )
        self.assertEqual(result, mock_response)

    @patch('ironicclient.common.base.Manager._list')
    def test_list_shards_api_version_mismatch(
        self, mock_list: mock.MagicMock,
    ) -> None:
        # Simulate a 404 error for an unsupported API version
        mock_list.side_effect = ValueError(
            "404 Not Found: The requested version is not supported"
        )

        # Initialize the ShardManager
        shard_manager = ShardManager(api=None)

        # Perform the test call and assert exception is raised
        with self.assertRaises(ValueError) as context:
            shard_manager.list(os_ironic_api_version="1.50")

        self.assertIn("404 Not Found", str(context.exception))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from osc_lib import utils as oscutils

from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


class ListBaremetalShard(command.Lister):
    """List baremetal shards."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalShard")

    def take_action(
        self,
        parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        data = client.shard.list()
        columns = res_fields.SHARD_RESOURCE.fields

        return (columns,
                (oscutils.get_item_properties(s, columns) for s in data))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import argparse
import importlib.metadata
import logging
import sys

from cliff import app
from cliff import commandmanager
import openstack
from openstack import config as os_config
from openstack.config import cloud_region as os_cloud_region
from osc_lib import utils

from ironicclient.common import http
from ironicclient.common.i18n import _
from ironicclient import exc
from ironicclient.v1 import client


_DEFAULTS: dict[str, str] = {
    'auth_type': 'none',
}
_TYPE: str = 'baremetal'
_DESCRIPTION: str = 'Bare Metal service (ironic) client'
_NAMESPACE: str = 'openstack.baremetal.v1'
_HELP: str = _("%(err)s.\n* Use --os-endpoint for standalone %(project)s.\n"
               "* Use --os-auth-url and credentials for authentication.\n"
               "* Use --os-cloud to load configuration from clouds.yaml\n"
               "* See `%(cmd)s --help` for more details")

LOG: logging.Logger = logging.getLogger(__name__)


class ClientManager(object):

    def __init__(
        self,
        cloud_region: os_cloud_region.CloudRegion,
        options: argparse.Namespace,
    ) -> None:
        self.cloud_region = cloud_region
        self.options = options
        self._ironic: client.Client | None = None

    @property
    def baremetal(self) -> client.Client:
        if self._ironic is None:
            self._ironic = self._create_ironic_client()
        return self._ironic

    def _create_ironic_client(self) -> client.Client:
        api_version = self.options.os_baremetal_api_version
        allow_api_version_downgrade = False
        if not api_version:
            api_version = self.cloud_region.get_default_microversion(_TYPE)
            if not api_version:
                api_version = http.LATEST_VERSION
                allow_api_version_downgrade = True
        LOG.debug(
            'Using bare metal API version %s, downgrade %s', api_version,
            'allowed' if allow_api_version_downgrade else 'disallowed')

        # NOTE(dtantsur): endpoint_override is required to respect settings in
        # clouds.yaml, such as baremetal_endpoint_override.
        endpoint_override = self.cloud_region.get_endpoint(_TYPE)
        try:
            return client.Client(
                os_ironic_api_version=api_version,
                allow_api_version_downgrade=allow_api_version_downgrade,
                session=self.cloud_region.get_session(),
                region_name=self.cloud_region.get_region_name(_TYPE),
                endpoint_override=endpoint_override,
                max_retries=self.options.max_retries,
                retry_interval=self.options.retry_interval,
            )
        except exc.EndpointNotFound as e:
            # Re-raise with a more obvious message.
            raise exc.EndpointNotFound(_HELP % {'err': e, 'cmd': sys.argv[0],
                                                'project': 'ironic'})


class CommandManager(commandmanager.CommandManager):

    def load_commands(self, namespace: str) -> None:
        super(CommandManager, self).load_commands(namespace)
        # Strip the 'baremetal' prefix used in OSC
        prefix = 'baremetal '
        prefix_len = len(prefix)
        self.commands = dict(
            (cmd[prefix_len:] if cmd.startswith(prefix) else cmd, ep)
            for (cmd, ep) in self.commands.items()
        )


class App(app.App):

    def __init__(self) -> None:
        version_info = importlib.metadata.version('python-ironicclient')
        mgr = CommandManager(_NAMESPACE)
        self.config = os_config.OpenStackConfig(override_defaults=_DEFAULTS)
        super(App, self).__init__(description=_DESCRIPTION,
                                  version=str(version_info),
                                  command_manager=mgr,
                                  deferred_help=True)

    def build_option_parser(
        self,
        description: str | None,
        version: str | None,
        argparse_kwargs: dict[str, object] | None = None,
    ) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super(App, self).build_option_parser(
            description, version, argparse_kwargs=argparse_kwargs)
        self.config.register_argparse_arguments(parser, sys.argv[1:])

        parser.add_argument(
            '--os-baremetal-api-version',
            metavar='<baremetal-api-version>',
            default=utils.env('OS_BAREMETAL_API_VERSION'),
            help='Bare metal API version, default="latest" (the maximum '
                 'version supported by both the client and the server). '
                 '(Env: OS_BAREMETAL_API_VERSION)',
        )
        parser.add_argument(
            '--max-retries',
            metavar='<max-retries-number>',
            default=http.DEFAULT_MAX_RETRIES,
            type=int,
            help='Maximum number of retries on connection problems and '
                 'resource state conflicts'
        )
        parser.add_argument(
            '--retry-interval',
            metavar='<retry-interval-seconds>',
            default=http.DEFAULT_RETRY_INTERVAL,
            type=int,
            help='Interval in seconds between two retries'
        )
        return parser

    # This is the openstacksdk default value
    _STREAM_FORMAT: str = '%(asctime)s %(levelname)s: %(name)s %(message)s'

    def _configure_ironic_logging(self) -> None:
        debug_enabled = self.options.debug or self.options.verbose_level > 1

        openstack.enable_logging(debug=debug_enabled, stream=sys.stderr,
                                 format_template=self._STREAM_FORMAT,
                                 # No fancy formatting if debug is off
                                 format_stream=debug_enabled)
        # NOTE(dtantsur): prevent openstack logging from appearing again in the
        # cliff handlers
        for name in ('openstack', 'keystoneauth'):
            logger = logging.getLogger(name)
            logger.propagate = False

        # NOTE(dtantsur): I wish logging.basicConfig worked, but at this point
        # cliff has already started configuring logging.
        logger = logging.getLogger('ironicclient')
        logger.setLevel(
            logging.DEBUG if debug_enabled else logging.WARNING)
        # warnings are already configured by cliff, only configure
        # debug logging for ironic.
        if not logger.handlers and debug_enabled:
            handler = logging.StreamHandler(stream=sys.stderr)
            handler.setFormatter(logging.Formatter(self._STREAM_FORMAT))
            logger.addHandler(handler)
            logger.propagate = False

    def initialize_app(self, argv: list[str]) -> None:
        super(App, self).initialize_app(argv)
        self._configure_ironic_logging()
        self.cloud_region = self.config.get_one(argparse=self.options)
        # Compatibility with OSC
        self.client_manager = ClientManager(self.cloud_region, self.options)


def main(argv: list[str] = sys.argv[1:]) -> int:
    result: int = App().run(argv)
    return result



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from ironicclient.common import base


class Shard(base.Resource):
    def __repr__(self) -> str:
        return "<Shard %s>" % self._info


class ShardManager(base.Manager[Shard]):
    resource_class: type[Shard] = Shard
    _resource_name: str = 'shards'

    def list(
        self,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Shard]:
        """Retrieve a list of shards.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of conductors.

        """
        return self._list(
            self._path(None),
            "shards",
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )
